(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2014, 2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2015)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>Abstract Completion\<close>

theory Abstract_Completion
imports
  "../Orderings/Reduction_Order"
  Peak_Decreasingness
begin

locale kb = reduction_order less
  for less :: "('a, 'b) term \<Rightarrow> ('a, 'b) term \<Rightarrow> bool" (infix "\<succ>" 50)
begin

inductive
  KB :: "('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> ('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> bool" (infix "\<turnstile>\<^sub>K\<^sub>B" 55)
where
  deduce: "(u, s) \<in> rstep R \<Longrightarrow> (u, t) \<in> rstep R \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B (E \<union> {(s, t)}, R)" |
  orientl: "s \<succ> t \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B (E - {(s, t)}, R \<union> {(s, t)})" |
  orientr: "t \<succ> s \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B (E - {(s, t)}, R \<union> {(t, s)})" |
  delete: "(s, s) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B (E - {(s, s)}, R)" |
  compose: "(t, u) \<in> rstep (R - {(s, t)}) \<Longrightarrow> (s, t) \<in> R \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B (E, (R - {(s, t)}) \<union> {(s, u)})" |
  simplifyl: "(s, u) \<in> rstep R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B ((E - {(s, t)}) \<union> {(u, t)}, R)" |
  simplifyr: "(t, u) \<in> rstep R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B ((E - {(s, t)}) \<union> {(s, u)}, R)" |
  collapse: "(t, u) \<in> rstep (R - {(t, s)}) \<Longrightarrow> (t, s) \<in> R \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B (E \<union> {(u, s)}, R - {(t, s)})"

lemma deduce_subset:
  assumes "E' = E \<union> {(s, t)}" and "R' = R"
    and "(u, s) \<in> rstep R" and "(u, t) \<in> rstep R"
  shows "E \<union> R \<subseteq> E' \<union> R'"
  using assms by fast

lemma deduce':
  assumes "E' = E \<union> {(s, t)}" and "R' = R"
    and "(u, s) \<in> rstep R" and "(u, t) \<in> rstep R"
  shows "E' \<union> R' \<subseteq> E \<union> R \<union> (rstep R)\<inverse> O rstep R"
  using assms by fast

lemma orientl_subset:
  assumes "E' = E - {(s, t)}" and "R' = R \<union> {(s, t)}"
    and "s \<succ> t" and "(s, t) \<in> E"
  shows "E \<union> R \<subseteq> E' \<union> R' \<union> R'\<inverse>"
  using assms by fast

lemma orientl':
  assumes "E' = E - {(s, t)}" and "R' = R \<union> {(s, t)}"
    and "s \<succ> t" and "(s, t) \<in> E"
  shows "E' \<union> R' \<subseteq> E \<union> R \<union> E\<inverse>"
  using assms by fast

lemma orientr_subset:
  assumes "E' = E - {(s, t)}" and "R' = R \<union> {(t, s)}"
    and "t \<succ> s" and "(s, t) \<in> E"
  shows "E \<union> R \<subseteq> E' \<union> R' \<union> R'\<inverse>"
  using assms by fast

lemma orientr':
  assumes "E' = E - {(s, t)}" and "R' = R \<union> {(t, s)}"
    and "t \<succ> s" and "(s, t) \<in> E"
  shows "E' \<union> R' \<subseteq> E \<union> R \<union> E\<inverse>"
  using assms by fast

lemma delete_subset:
  assumes "E' = E - {(s, s)}" and "R' = R"
    and "(s, s) \<in> E"
  shows "E \<union> R \<subseteq> E' \<union> R' \<union> Id"
  using assms by fast

lemma delete':
  assumes "E' = E - {(s, s)}" and "R' = R"
    and "(s, s) \<in> E"
  shows "E' \<union> R' \<subseteq> E \<union> R"
  using assms by fast

lemma compose_subset:
  assumes "E' = E" and "R' = (R - {(s, t)}) \<union> {(s, u)}"
    and "(t, u) \<in> rstep (R - {(s, t)})" and "(s, t) \<in> R"
  shows "E \<union> R \<subseteq> E' \<union> R' \<union> rstep R' O (rstep R')\<inverse>"
  using assms by fast

lemma compose':
  assumes "E' = E" and "R' = (R - {(s, t)}) \<union> {(s, u)}"
    and "(t, u) \<in> rstep (R - {(s, t)})" and "(s, t) \<in> R"
  shows "E' \<union> R' \<subseteq> E \<union> R \<union> rstep R O rstep R"
  using assms by fast

lemma simplifyl_subset:
  assumes "E' = (E - {(s, t)}) \<union> {(u, t)}" and "R' = R"
    and "(s, u) \<in> rstep R" and "(s, t) \<in> E"
  shows "E \<union> R \<subseteq> E' \<union> R' \<union> rstep R' O rstep E' \<union> rstep E' O (rstep R')\<inverse>"
  using assms by fast

lemma simplifyl':
  assumes "E' = (E - {(s, t)}) \<union> {(u, t)}" and "R' = R"
    and "(s, u) \<in> rstep R" and "(s, t) \<in> E"
  shows "E' \<union> R' \<subseteq> E \<union> R \<union> (rstep R)\<inverse> O rstep E \<union> rstep E O rstep R"
  using assms by fast

lemma simplifyr_subset:
  assumes "E' = (E - {(s, t)}) \<union> {(s, u)}" and "R' = R"
    and "(t, u) \<in> rstep R" and "(s, t) \<in> E"
  shows "E \<union> R \<subseteq> E' \<union> R' \<union> rstep R' O rstep E' \<union> rstep E' O (rstep R')\<inverse>"
  using assms by fast

lemma simplifyr':
  assumes "E' = (E - {(s, t)}) \<union> {(s, u)}" and "R' = R"
    and "(t, u) \<in> rstep R" and "(s, t) \<in> E"
  shows "E' \<union> R' \<subseteq> E \<union> R \<union> (rstep R)\<inverse> O rstep E \<union> rstep E O rstep R"
  using assms by fast

lemma collapse_subset:
  assumes "E' = E \<union> {(u, s)}" and "R' = R - {(t, s)}"
    and "(t, u) \<in> rstep (R - {(t, s)})" and "(t, s) \<in> R"
  shows "E \<union> R \<subseteq> E' \<union> R' \<union> rstep R' O rstep E'"
  using assms by fast

lemma collapse':
  assumes "E' = E \<union> {(u, s)}" and "R' = R - {(t, s)}"
    and "(t, u) \<in> rstep (R - {(t, s)})" and "(t, s) \<in> R"
  shows "E' \<union> R' \<subseteq> E \<union> R \<union> (rstep R)\<inverse> O rstep R"
  using assms by fast

lemma rstep_reflcl_symcl_Un_idemp [simp]:
  "rstep (((rstep R \<union> rstep S)\<^sup>\<leftrightarrow>)\<^sup>*) = ((rstep R \<union> rstep S)\<^sup>\<leftrightarrow>)\<^sup>*"
  unfolding symcl_Un
  unfolding rstep_simps(5) [symmetric]
  unfolding rstep_union [symmetric]
  unfolding rstep_rtrancl_idemp
  by (rule refl)

lemma KB_subset:
  assumes "(E, R) \<turnstile>\<^sub>K\<^sub>B (E', R')"
  shows "rstep (E \<union> R) \<subseteq> (rstep R')\<^sup>= O (rstep (E' \<union> R'))\<^sup>= O ((rstep R')\<inverse>)\<^sup>="
proof -
  have "E \<union> R \<subseteq> (rstep R')\<^sup>= O (rstep (E' \<union> R'))\<^sup>= O ((rstep R')\<inverse>)\<^sup>=" (is "?L \<subseteq> ?R")
  proof
    fix s t
    assume *: "(s, t) \<in> ?L"
    from assms show "(s, t) \<in> ?R"
    proof (cases)
      case deduce
      from deduce_subset [OF this] show ?thesis using * by blast
    next
      case orientl
      from orientl_subset [OF this] show ?thesis using * by blast
    next
      case orientr
      from orientr_subset [OF this] show ?thesis using * by blast
    next
      case delete
      from delete_subset [OF this] show ?thesis using * by blast
    next
      case compose
      from compose_subset [OF this] show ?thesis using * by blast
    next
      case simplifyl
      from simplifyl_subset [OF this] show ?thesis using * by blast
    next
      case simplifyr
      from simplifyr_subset [OF this] show ?thesis using * by blast
    next
      case collapse
      from collapse_subset [OF this] show ?thesis using * by blast
    qed
  qed
  from rstep_mono [OF this] show ?thesis by (simp add: rstep_simps)
qed

lemma KB_subset':
  assumes "(E, R) \<turnstile>\<^sub>K\<^sub>B (E', R')"
  shows "rstep (E' \<union> R') \<subseteq> ((rstep (E \<union> R))\<^sup>\<leftrightarrow>)\<^sup>*"
    (is "?L \<subseteq> ?R")
proof -
  have "E' \<union> R' \<subseteq> ((rstep (E \<union> R))\<^sup>\<leftrightarrow>)\<^sup>*" (is "?L \<subseteq> ?R")
  proof
    fix s t
    assume *: "(s, t) \<in> ?L"
    from assms show "(s, t) \<in> ?R"
    proof (cases)
      case deduce
      from deduce' [OF this] and *
        have "(s, t) \<in> E \<union> R \<or> (s, t) \<in> (rstep R)\<inverse> O rstep R" by auto
      then show ?thesis
      proof
        assume "(s, t) \<in> (rstep R)\<inverse> O rstep R"
        then obtain u where "(s, u) \<in> (rstep R)\<inverse>" and "(u, t) \<in> rstep R" by auto
        then have "(s, u) \<in> ?R" and "(u, t) \<in> ?R" by auto
        then show ?thesis by simp
      qed blast
    next
      case orientl
      from orientl' [OF this] show ?thesis using * by blast
    next
      case orientr
      from orientr' [OF this] show ?thesis using * by blast
    next
      case delete
      from delete' [OF this] show ?thesis using * by blast
    next
      case compose
      from compose' [OF this] and *
        have "(s, t) \<in> E \<union> R \<or> (s, t) \<in> rstep R O rstep R" by auto
      then show ?thesis
      proof
        assume "(s, t) \<in> rstep R O rstep R"
        then obtain u where "(s, u) \<in> rstep R" and "(u, t) \<in> rstep R" by auto
        then have "(s, u) \<in> ?R" and "(u, t) \<in> ?R" by auto
        then show ?thesis by simp
      qed blast
    next
      case simplifyl
      from simplifyl' [OF this] and *
        have "(s, t) \<in> E \<union> R \<or> (s, t) \<in> (rstep R)\<inverse> O rstep E \<or> (s, t) \<in> rstep E O rstep R" by auto
      then show ?thesis
      proof (elim disjE)
        assume "(s, t) \<in> (rstep R)\<inverse> O rstep E"
        then obtain u where "(s, u) \<in> (rstep R)\<inverse>" and "(u, t) \<in> rstep E" by auto
        then have "(s, u) \<in> ?R" and "(u, t) \<in> ?R" by auto
        then show ?thesis by simp
      next
        assume "(s, t) \<in> rstep E O rstep R"
        then obtain u where "(s, u) \<in> rstep E" and "(u, t) \<in> rstep R" by auto
        then have "(s, u) \<in> ?R" and "(u, t) \<in> ?R" by auto
        then show ?thesis by simp
      qed blast
    next
      case simplifyr
      from simplifyr' [OF this] and *
        have "(s, t) \<in> E \<union> R \<or> (s, t) \<in> (rstep R)\<inverse> O rstep E \<or> (s, t) \<in> rstep E O rstep R" by auto
      then show ?thesis
      proof (elim disjE)
        assume "(s, t) \<in> (rstep R)\<inverse> O rstep E"
        then obtain u where "(s, u) \<in> (rstep R)\<inverse>" and "(u, t) \<in> rstep E" by auto
        then have "(s, u) \<in> ?R" and "(u, t) \<in> ?R" by auto
        then show ?thesis by simp
      next
        assume "(s, t) \<in> rstep E O rstep R"
        then obtain u where "(s, u) \<in> rstep E" and "(u, t) \<in> rstep R" by auto
        then have "(s, u) \<in> ?R" and "(u, t) \<in> ?R" by auto
        then show ?thesis by simp
      qed blast
    next
      case collapse
      from collapse' [OF this] and *
        have "(s, t) \<in> E \<union> R \<or> (s, t) \<in> (rstep R)\<inverse> O rstep R" by auto
      then show ?thesis
      proof
        assume "(s, t) \<in> (rstep R)\<inverse> O rstep R"
        then obtain u where "(s, u) \<in> (rstep R)\<inverse>" and "(u, t) \<in> rstep R" by auto
        then have "(s, u) \<in> ?R" and "(u, t) \<in> ?R" by auto
        then show ?thesis by auto
      qed blast
    qed
  qed
  from rstep_mono [OF this] show ?thesis by (simp add: rstep_simps)
qed

lemma step_subset:
  "(rstep R')\<^sup>= O (rstep (E' \<union> R'))\<^sup>= O ((rstep R')\<inverse>)\<^sup>= \<subseteq> ((rstep (E' \<union> R'))\<^sup>\<leftrightarrow>)\<^sup>*"
  (is "?L \<subseteq> ?R")
proof
  fix s t assume "(s, t) \<in> ?L"
  then obtain u and v where "(s, u) \<in> (rstep R')\<^sup>="
    and "(u, v) \<in> (rstep (E' \<union> R'))\<^sup>="
    and "(v, t) \<in> ((rstep R')\<inverse>)\<^sup>=" by auto
  then have "(s, u) \<in> ?R" and "(u, v) \<in> ?R" and "(v, t) \<in> ?R" by auto
  then show "(s, t) \<in> ?R" by auto
qed

lemma KB_conversion:
  assumes "(E, R) \<turnstile>\<^sub>K\<^sub>B (E', R')"
  shows "(rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*" (is "?L = ?R")
proof
  note * = subset_trans [OF KB_subset [OF assms] step_subset]
  with converse_mono [THEN iffD2, OF *]
    have "(rstep (E \<union> R))\<^sup>\<leftrightarrow> \<subseteq> ?R"
      unfolding conversion_def
      unfolding rtrancl_converse [symmetric]
      unfolding symcl_converse by blast
  from rtrancl_mono [OF this] show "?L \<subseteq> ?R"
    unfolding conversion_def and rtrancl_idemp .
next
  note * = KB_subset' [OF assms]
  with converse_mono [THEN iffD2, OF *]
    have "(rstep (E' \<union> R'))\<^sup>\<leftrightarrow> \<subseteq> ?L"
      unfolding conversion_def
      unfolding rtrancl_converse [symmetric]
      unfolding symcl_converse by blast
  from rtrancl_mono [OF this] show "?R \<subseteq> ?L"
    unfolding conversion_def and rtrancl_idemp .
qed

lemma KB_rtrancl_conversion:
  assumes "KB\<^sup>*\<^sup>* (E, R) (E', R')"
  shows "(rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using assms
  by (induct "(E, R)" "(E', R')" arbitrary: E' R')
     (force dest: KB_conversion)+

lemma KB_rtrancl_rules_subset_less:
  assumes "KB\<^sup>*\<^sup>* (E, R) (E', R')" and "R \<subseteq> {(x, y). x \<succ> y}"
  shows "R' \<subseteq> {(x, y). x \<succ> y}"
using assms
proof (induction "(E, R)" "(E', R')" arbitrary: E' R')
  case (rtrancl_into_rtrancl ER'')
  moreover then obtain E'' and R''
    where [simp]: "ER'' = (E'', R'')" by force
  ultimately have "(E'', R'') \<turnstile>\<^sub>K\<^sub>B (E', R')"
    and "R'' \<subseteq> {(x, y). x \<succ> y}" by simp+
  moreover
  {
    fix s t u
    assume "(t, u) \<in> rstep (R'' - {(s, t)})" and "(s, t) \<in> R''" and "R'' \<subseteq> {(x, y). x \<succ> y}"
    then have "s \<succ> u" by (induct t u) (blast intro: subst ctxt dest: trans)
  }
  ultimately show ?case by (cases) auto
qed

definition mstep :: "('a, 'b) term multiset \<Rightarrow> ('a, 'b) trs \<Rightarrow> ('a, 'b) term rel" where
  "mstep M R = {(s, t). (s, t) \<in> rstep R \<and> (\<exists>s' t'. s' \<in># M \<and> t' \<in># M \<and> s' \<succeq> s \<and> t' \<succeq> t)}"

lemma mstep_iff:
  "(x, y) \<in> mstep M R \<longleftrightarrow> (x, y) \<in> rstep R \<and> (\<exists>s' t'. s' \<in># M \<and> t' \<in># M \<and> s' \<succeq> x \<and> t' \<succeq> y)"
  by (auto simp: mstep_def)

lemma mstep_Un [simp]:
  "mstep M (R \<union> R') = mstep M R \<union> mstep M R'"
  by (auto iff: mstep_iff)

lemma step_subset':
  "(mstep M R')\<^sup>= O (mstep M (E' \<union> R'))\<^sup>= O ((mstep M R')\<inverse>)\<^sup>= \<subseteq> ((mstep M (E' \<union> R'))\<^sup>\<leftrightarrow>)\<^sup>*"
  (is "?L \<subseteq> ?R")
proof
  fix s t assume "(s, t) \<in> ?L"
  then obtain u and v where "(s, u) \<in> (mstep M R')\<^sup>="
    and "(u, v) \<in> (mstep M (E' \<union> R'))\<^sup>="
    and "(v, t) \<in> ((mstep M R')\<inverse>)\<^sup>=" by auto
  then have "(s, u) \<in> ?R" and "(u, v) \<in> ?R" and "(v, t) \<in> ?R" by auto
  then show "(s, t) \<in> ?R" by auto
qed

lemma rstep_subset_less:
  assumes "R \<subseteq> {(x, y). x \<succ> y}"
  shows "rstep R \<subseteq> {(x, y). x \<succ> y}"
proof
  fix x y assume "(x, y) \<in> rstep R" then show "(x, y) \<in> {(x, y). x \<succ> y}"
    using assms by (induct) (auto intro: subst ctxt)
qed

lemma rsteps_subset_less:
  assumes "R \<subseteq> {(x, y). x \<succ> y}"
  shows "(rstep R)\<^sup>+ \<subseteq> {(x, y). x \<succ> y}"
proof
  fix s t
  assume "(s, t) \<in> (rstep R)\<^sup>+"
  then show "(s, t) \<in> {(x, y). x \<succ> y}"
  proof (induct)
    case (base u)
    with rstep_subset_less [OF assms] show ?case by auto
  next
    case (step t u)
    then show ?case
      using rstep_subset_less [OF assms]
      by auto (metis in_mono mem_Collect_eq split_conv trans)
  qed
qed

lemma mstep_subset:
  assumes "(E, R) \<turnstile>\<^sub>K\<^sub>B (E', R')" and "R' \<subseteq> {(x, y). x \<succ> y}"
  shows "mstep M (E \<union> R) \<subseteq> (mstep M (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
proof
  fix s t
  assume "(s, t) \<in> mstep M (E \<union> R)"
  then obtain s' and t' where "s' \<in># M" and "t' \<in># M"
    and "s' \<succeq> s" and "t' \<succeq> t" and "(s, t) \<in> rstep (E \<union> R)"
    by (auto simp: mstep_def)
  with KB_subset [OF assms(1)] obtain u and v
    where "(s, u) \<in> (rstep R')\<^sup>=" and "(u, v) \<in> (rstep (E' \<union> R'))\<^sup>="
    and "(v, t) \<in> ((rstep R')\<inverse>)\<^sup>=" by blast
  moreover then have "(t, v) \<in> (rstep R')\<^sup>=" by auto
  ultimately have "s \<succeq> u" and "t \<succeq> v" using rstep_subset_less [OF assms(2)] by auto
  with \<open>s' \<succeq> s\<close> and \<open>t' \<succeq> t\<close> have "s' \<succeq> u" and "t' \<succeq> v" using trans by blast+
  then have "(s, u) \<in> (mstep M R')\<^sup>="
    and "(u, v) \<in> (mstep M (E' \<union> R'))\<^sup>="
    and "(v, t) \<in> ((mstep M R')\<inverse>)\<^sup>="
    using \<open>s' \<in># M\<close> and \<open>t' \<in># M\<close>
    and \<open>s' \<succeq> s\<close> and \<open>t' \<succeq> t\<close>
    and \<open>(s, u) \<in> (rstep R')\<^sup>=\<close>
    and \<open>(u, v) \<in> (rstep (E' \<union> R'))\<^sup>=\<close>
    and \<open>(v, t) \<in> ((rstep R')\<inverse>)\<^sup>=\<close>
    unfolding mstep_def by (blast)+
  then have "(s, t) \<in> (mstep M R')\<^sup>= O ((mstep M (E' \<union> R')))\<^sup>= O ((mstep M R')\<inverse>)\<^sup>=" by blast
  with step_subset' [of M R' E'] show "(s, t) \<in> (mstep M (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
    unfolding conversion_def by blast
qed

lemma mstep_subset':
  assumes "(E, R) \<turnstile>\<^sub>K\<^sub>B (E', R')" and "R' \<subseteq> {(x, y). x \<succ> y}"
  shows "(mstep M (E \<union> R))\<inverse> \<subseteq> (mstep M (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using converse_mono [THEN iffD2, OF mstep_subset [OF assms, of M]]
  by simp

lemma mstep_symcl_subset:
  assumes "(E, R) \<turnstile>\<^sub>K\<^sub>B (E', R')" and "R' \<subseteq> {(x, y). x \<succ> y}"
  shows "(mstep M (E \<union> R))\<^sup>\<leftrightarrow> \<subseteq> (mstep M (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using mstep_subset [OF assms] and mstep_subset' [OF assms] by blast

lemma msteps_subset:
  assumes "(E, R) \<turnstile>\<^sub>K\<^sub>B (E', R')" and "R' \<subseteq> {(x, y). x \<succ> y}"
  shows "(mstep M (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (mstep M (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using mstep_symcl_subset [OF assms, THEN rtrancl_mono] by (simp add: conversion_def)

lemma UNIV_mstep_rstep_iff [simp]:
  "(\<Union>M\<in>UNIV. mstep M R) = rstep R"
proof -
  have *: "\<And>s t. (s, t) \<in> rstep R \<Longrightarrow>
    s \<in># {#s, t#} \<and> t \<in># {#s, t#} \<and> s \<succeq> s \<and> t \<succeq> t \<and>
  (s, t) \<in> mstep {#s, t#} R" by (auto iff: mstep_iff)
  show ?thesis by (auto iff: mstep_iff) (insert *, blast)
qed

lemma rstep_imp_mstep:
  assumes "(t, u) \<in> rstep R" and "s \<in># M" and "s \<succeq> t" and "t \<succ> u"
  shows "(t, u) \<in> mstep M R"
  using assms by (auto simp: mstep_iff dest: trans)

lemma rsteps_imp_msteps:
  assumes "t \<in># M" and "(t, u) \<in> (rstep R)\<^sup>*" and "R \<subseteq> {(x, y). x \<succ> y}"
  shows "(t, u) \<in> (mstep M R)\<^sup>*"
  using assms(2, 1)
proof (induct)
  case base show ?case by simp
next
  note less = rstep_subset_less [OF assms(3)]
  case (step u v)
  have "(u, v) \<in> rstep R" by fact
  moreover have "t \<in># M" by fact
  moreover
  from \<open>(t, u) \<in> (rstep R)\<^sup>*\<close> and less
    have "t \<succeq> u" by (induct) (auto dest: trans)
  moreover have "u \<succ> v" using less and step by blast
  ultimately have "(u, v) \<in> mstep M R" by (rule rstep_imp_mstep)
  with step show ?case by auto
qed

lemma finite_runD:
  assumes "R 0 = {}" and "E n = {}"
    and run: "\<forall>i<n. (E i, R i) \<turnstile>\<^sub>K\<^sub>B (E (Suc i), R (Suc i))"
  shows finite_run_imp_conversion_eq: "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>*"
    and finite_run_imp_SN: "SN (rstep (R n))"
    and finite_run_imp_R_less: "\<And>i. i \<le> n \<Longrightarrow> R i \<subseteq> {(x, y). x \<succ> y}"
proof -
  have *: "\<And>i. i \<le> n \<Longrightarrow> KB\<^sup>*\<^sup>* (E 0, R 0) (E i, R i)"
  proof -
    fix i assume "i \<le> n"
    with run show "KB\<^sup>*\<^sup>* (E 0, R 0) (E i, R i)"
      by (induct i) (auto, metis (hide_lams, no_types) Suc_le_eq rtranclp.rtrancl_into_rtrancl)
  qed
  from * [THEN KB_rtrancl_rules_subset_less] and \<open>R 0 = {}\<close>
    show "\<And>i. i \<le> n \<Longrightarrow> R i \<subseteq> {(x, y). x \<succ> y}" by auto
  with rstep_subset_less [of "R n"]
    show "SN (rstep (R n))" using SN_less by (simp add: assms) (metis SN_subset)
  from KB_rtrancl_conversion [OF * [of n]]
    show "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>*" by (simp add: assms)
qed

lemma finite_run:
  assumes "R 0 = {}" and "E n = {}"
    and run: "\<forall>i<n. (E i, R i) \<turnstile>\<^sub>K\<^sub>B (E (Suc i), R (Suc i))"
    (*the peak-condition is derivable from the critical pair lemma + fairness*)
    and peak: "(rstep (R n))\<inverse> O rstep (R n) \<subseteq> (rstep (R n))\<^sup>\<down> \<union> (\<Union>i\<le>n. (rstep (E i))\<^sup>\<leftrightarrow>)"
  shows "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>* \<and> SN (rstep (R n)) \<and> CR (rstep (R n))"
proof -
  have R_less: "\<And>i. i \<le> n \<Longrightarrow> R i \<subseteq> {(x, y). x \<succ> y}"
    and "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>*"
    and "SN (rstep (R n))"
    by (fact finite_runD [OF assms(1-3)])+
  moreover
  have "CR (rstep (R n))"
  proof -
    def \<R> \<equiv> "\<Union>M\<in>UNIV. mstep M (R n)"
    def mless \<equiv> "\<lambda>x y. mulex less\<inverse>\<inverse> y x"
    write mless (infix "\<succ>''" 50)
    interpret lab: ars_labeled_sn \<R> "\<lambda>M. mstep M (R n)" UNIV "op \<succ>'"
    proof
      show "\<R> = (\<Union>M. mstep M (R n))" using \<R>_def R_less by simp
      show "SN {(x, y). x \<succ>' y}" unfolding mless_def by (metis SN_less SN_mulex)
    qed
    interpret ars_peak_decreasing \<R> "\<lambda>M. mstep M (R n)" UNIV "op \<succ>'"
    proof (rule lab.ars_peak_decreasing_alt_intro)
      fix s t u :: "('a, 'b) term" and M\<^sub>1 M\<^sub>2 :: "('a, 'b) term multiset"
      def M \<equiv> "{#t, u#}"
      assume "(s, t) \<in> mstep M\<^sub>1 (R n)" and "(s, u) \<in> mstep M\<^sub>2 (R n)"
      then have "(t, s) \<in> (rstep (R n))\<inverse>" and "(s, u) \<in> rstep (R n)" by (auto iff: mstep_iff)
      then have "(t, u) \<in> (rstep (R n))\<inverse> O rstep (R n)" by blast
      with peak have "(t, u) \<in> (rstep (R n))\<^sup>\<down> \<or> (t, u) \<in> (\<Union>i\<le>n. (rstep (E i))\<^sup>\<leftrightarrow>)" by auto
      then have "(t, u) \<in> (mstep M (R n))\<^sup>\<leftrightarrow>\<^sup>*"
      proof
        assume "(t, u) \<in> (rstep (R n))\<^sup>\<down>"
        then obtain v where 1: "(t, v) \<in> (rstep (R n))\<^sup>*"
          and 2: "(u, v) \<in> (rstep (R n))\<^sup>*" by auto
        from rsteps_imp_msteps [of M, OF _ 1 R_less]
          have "(t, v) \<in> (mstep M (R n))\<^sup>\<leftrightarrow>\<^sup>*" by (auto simp: M_def)
        moreover
        from rsteps_imp_msteps [of M, OF _ 2 R_less]
          have "(v, u) \<in> (mstep M (R n))\<^sup>\<leftrightarrow>\<^sup>*" by (auto simp: M_def) (metis conversion_def in_rtrancl_UnI rtrancl_converseI)
        ultimately show ?thesis by (auto simp: conversion_def)
      next
        assume "(t, u) \<in> (\<Union>i\<le>n. (rstep (E i))\<^sup>\<leftrightarrow>)"
        then obtain i where "i \<le> n" and "(t, u) \<in> (rstep (E i))\<^sup>\<leftrightarrow>" by auto
        then have "(t, u) \<in> (mstep M (E i))\<^sup>\<leftrightarrow>" by (auto iff: mstep_iff simp: M_def)
        then have "(t, u) \<in> (mstep M (E i \<union> R i))\<^sup>\<leftrightarrow>\<^sup>*" by (auto simp: conversion_def)
        then show ?thesis using \<open>i \<le> n\<close>
        proof (induct "n - i" arbitrary: i)
          case 0
          with \<open>i \<le> n\<close> have [simp]: "i = n" by arith
          with 0 have "(t, u) \<in> (mstep M (R n))\<^sup>\<leftrightarrow>\<^sup>*" by (simp add: assms)
          then show ?case by auto
        next
          case (Suc m)
          from \<open>Suc m = n - i\<close> have *: "m = n - Suc i" and leq: "Suc i \<le> n" by auto
          from run have **: "(E i, R i) \<turnstile>\<^sub>K\<^sub>B (E (Suc i), R (Suc i))" using leq by auto
          from \<open>(t, u) \<in> (mstep M (E i \<union> R i))\<^sup>\<leftrightarrow>\<^sup>*\<close>
            have "(t, u) \<in> (mstep M (E (Suc i) \<union> R (Suc i)))\<^sup>\<leftrightarrow>\<^sup>*"
            using msteps_subset [OF ** R_less [OF leq]] by blast
          from Suc.hyps(1) [OF * this leq]
            show ?case .
        qed
      qed
      moreover have "M \<in> UNIV" by auto
      moreover have "M\<^sub>1 \<succ>' M"
      proof -
        from \<open>(s, t) \<in> mstep M\<^sub>1 (R n)\<close> have ne1: "M\<^sub>1 \<noteq> {#}" by (auto iff: mstep_iff)
        have "\<forall>u. u \<in># M \<longrightarrow> (\<exists>v. v \<in># M\<^sub>1 \<and> v \<succ> u)"
        proof -
          { have "\<exists>v. v \<in># M\<^sub>1 \<and> v \<succ> t"
              using \<open>(s, t) \<in> mstep M\<^sub>1 (R n)\<close>
              and rstep_subset_less [OF R_less [of n]]
              by (auto simp: mstep_iff dest: trans) }
          moreover
          { have "\<exists>v. v \<in># M\<^sub>1 \<and> v \<succ> u"
              using \<open>(s, t) \<in> mstep M\<^sub>1 (R n)\<close> and \<open>(s, u) \<in> mstep M\<^sub>2 (R n)\<close>
              and rstep_subset_less [OF R_less [of n]] by (auto simp: mstep_iff dest: trans) }
          ultimately show ?thesis by (auto simp: M_def)
        qed
        from mulex_on_all_strict [OF ne1 _ _ this, of UNIV]
          have less1: "mulex less\<inverse>\<inverse> M M\<^sub>1" by auto (metis (lifting, no_types) conversep.intros mulex_on_mono)
        then show ?thesis by (simp add: mless_def)
      qed
      ultimately show "\<exists>M. M \<in> UNIV \<and> (M\<^sub>1 \<succ>' M \<or> M\<^sub>2 \<succ>' M) \<and> (t, u) \<in> (mstep M (R n))\<^sup>\<leftrightarrow>\<^sup>*" by blast
    qed
    from CR show "CR (rstep (R n))" by (simp add: \<R>_def)
  qed
  ultimately show ?thesis by blast
qed

end

end

