theory Entailment_Checker
imports
  ITS
  "~~/src/HOL/Library/Product_Lexorder"
  "$AFP/Check_Monad/Check_Monad"
begin

fun scalar_mult :: "int \<Rightarrow> 'v lin_exp \<Rightarrow> 'v lin_exp" where
  "scalar_mult d (Lin_Exp c cxs) = (Lin_Exp (d * c) (map (\<lambda> (x,c). (x, d * c)) cxs))"

lemma scalar_mult[simp]: "eval_lin_exp \<alpha> (scalar_mult d e) = d * eval_lin_exp \<alpha> e"
proof (cases e)
  case (Lin_Exp c cxs)
  show ?thesis unfolding Lin_Exp
    by (simp add: field_simps o_def, induct cxs, auto simp: field_simps)
qed

fun add_lin_exps :: "'v lin_exp list \<Rightarrow> 'v lin_exp" where
  "add_lin_exps [] = Lin_Exp 0 []"
| "add_lin_exps (Lin_Exp c cxs # exps) = (case add_lin_exps exps of Lin_Exp d dxs
    \<Rightarrow> Lin_Exp (c + d) (cxs @ dxs))"

lemma add_lin_exps[simp]: "eval_lin_exp \<alpha> (add_lin_exps exps) = listsum (map (eval_lin_exp \<alpha>) exps)"
proof (induct exps rule: add_lin_exps.induct)
  case (2 c cxs exps)
  obtain d dxs where id: "add_lin_exps exps = Lin_Exp d dxs" by (cases "add_lin_exps exps", auto)
  from 2 show ?case by (simp add: id)
qed simp

fun normalize_sum_duplicates :: "('v \<times> int) list \<Rightarrow> ('v \<times> int) list" where
  "normalize_sum_duplicates [] = []"
| "normalize_sum_duplicates ((x,c) # cxs) = (case partition (\<lambda> (y,d). y = x) cxs of
     (xs, dys) \<Rightarrow> (x,c + listsum (map snd xs)) # normalize_sum_duplicates dys)"

lemma partition_listsum: "partition f xs = (yes,no) \<Longrightarrow> listsum (map g xs) = 
  listsum (map g yes) + (listsum (map g no) :: 'a :: ab_group_add)"
proof (induct xs arbitrary: yes no)
  case (Cons x xs)
  obtain y n where p: "partition f xs = (y,n)" by force
  from Cons(2)[unfolded partition.simps p Let_def split]
  have id: "(yes, no) = (if f x then (x # y, n) else (y, x # n))" by auto
  from Cons(1)[OF p] have IH: "listsum (map g xs) = listsum (map g y) + listsum (map g n)" .
  show ?case using id by (cases "f x", auto simp: IH ac_simps)
qed simp

lemma normalize_sum_duplicates[simp]: 
  "(\<Sum>(x, c)\<leftarrow> normalize_sum_duplicates cxs. c * \<alpha> x) = (\<Sum>(x, c) \<leftarrow> cxs. c * \<alpha> x)"
proof (induct cxs rule: normalize_sum_duplicates.induct)
  case (2 x c cxs)
  obtain xs dys where part: "partition (\<lambda> (y,d). y = x) cxs = (xs,dys)" by force
  hence xs: "\<And> y c. (y,c) \<in> set xs \<Longrightarrow> y = x" by auto
  show ?case unfolding normalize_sum_duplicates.simps part split
    by (simp add: field_simps partition_listsum[OF part] 2[OF part[symmetric]],
    insert xs, induct xs, (force simp: field_simps)+)
qed simp

lemma normalize_sum_duplicates_distinct: 
  "distinct (map fst (normalize_sum_duplicates cxs)) \<and> fst ` set (normalize_sum_duplicates cxs) = fst ` set cxs"
proof (induct cxs rule: normalize_sum_duplicates.induct)
  case (2 x c cxs)
  obtain xs dys where part: "partition (\<lambda> (y,d). y = x) cxs = (xs,dys)" by force
  hence dys: "\<And> y c. (y,c) \<in> set dys \<Longrightarrow> y \<noteq> x" by auto
  show ?case using 2[OF part[symmetric]] unfolding normalize_sum_duplicates.simps part split
    using part by auto
qed simp

definition normalize_sum_0 :: "('v \<times> int)list \<Rightarrow> ('v \<times> int)list" where
  "normalize_sum_0 = filter (\<lambda> (x,c). c \<noteq> 0)"

lemma normalize_sum_0: "(\<Sum>(x, c)\<leftarrow> normalize_sum_0 cxs. c * \<alpha> x) = (\<Sum>(x, c) \<leftarrow> cxs. c * \<alpha> x)"
  unfolding normalize_sum_0_def by (induct cxs, auto)

definition normalize_sum :: "('v :: linorder \<times> int) list \<Rightarrow> ('v \<times> int) list" where
  "normalize_sum = normalize_sum_0 o sort o normalize_sum_duplicates"

lemma normalize_sum: 
  "(\<Sum>(x, c)\<leftarrow> normalize_sum cxs. c * \<alpha> x) = (\<Sum>(x, c) \<leftarrow> cxs. c * \<alpha> x)"
proof -
  from normalize_sum_duplicates_distinct[of cxs]
  have dist: "distinct (normalize_sum_duplicates cxs)"
    by (simp add: distinct_map)
  show ?thesis
    unfolding normalize_sum_def o_def normalize_sum_0
    by (subst setsum.distinct_set_conv_list[symmetric], simp add: dist, simp,
    subst setsum.distinct_set_conv_list[OF dist], simp)
qed

context
  fixes ty :: "'v :: {show,linorder}itself"
begin

fun trivial_entailment_checker :: "'v lin_exp \<Rightarrow> 'v lin_exp \<Rightarrow> shows check" where
  "trivial_entailment_checker (Lin_Exp c exp1) (Lin_Exp d exp2) = (let
     nexp1 = normalize_sum exp1;
     nexp2 = normalize_sum exp2
   in
     check (nexp1 = nexp2 \<and> c \<ge> d \<or> c > 0 \<and> nexp1 = []) (shows ''could not ensure trivial entailment\<newline>  ''
     o shows_lin_atom (Exp_Le_0 (Lin_Exp c nexp1)) o shows '' ==> '' o shows_lin_atom (Exp_Le_0 (Lin_Exp d nexp2))))"

lemma trivial_entailment_checker: assumes ok: "isOK(trivial_entailment_checker exp1 exp2)"
  shows "Conjunction [Exp_Le_0 exp1] implies Conjunction [Exp_Le_0 exp2]"
proof -
  obtain c1 cxs1 where exp1: "exp1 = Lin_Exp c1 cxs1" by (cases exp1, auto)
  obtain c2 cxs2 where exp2: "exp2 = Lin_Exp c2 cxs2" by (cases exp2, auto)
  let ?n1 = "normalize_sum cxs1" let ?n2 = "normalize_sum cxs2"
  from ok[unfolded exp1 exp2] have choice: "?n1 = ?n2 \<and> c2 \<le> c1 \<or> c1 > 0 \<and> ?n1 = []" by auto
  thus ?thesis unfolding exp1 exp2
    by (standard, auto simp: entailment_def normalize_sum[symmetric, of _ cxs1] normalize_sum)
qed

definition entailment_checker_atom :: "nat list \<Rightarrow> 'v formula \<Rightarrow> 'v lin_atom \<Rightarrow> shows check" where
  "entailment_checker_atom lambdas \<phi> \<psi> \<equiv>
    let linexp = add_lin_exps (map (\<lambda> (c,e). scalar_mult (int c) e) (zip lambdas (map lin_exp (atoms \<phi>))))
    in trivial_entailment_checker linexp (lin_exp \<psi>)
    <+? (\<lambda> s. shows ''could not ensure entailment of atom\<newline>  '' o 
      shows_formula \<phi> o shows '' ==> '' o shows_lin_atom \<psi> o
      shows ''\<newline>since after applying hints the following error occurred\<newline>'' o s)"

lemma listsum_le_0: assumes "\<And> x. x \<in> set xs \<Longrightarrow> x \<le> 0"
  shows "listsum xs \<le> (0 :: int)"
  using assms by (induct xs, fastforce+)

lemma eval_lin_exp[simp]: "eval_lin_exp \<alpha> (lin_exp x) \<le> 0 \<longleftrightarrow> eval_lin_atom \<alpha> x"
  by (cases x, auto)

lemma entailment_checker_atom: assumes ok: "isOK(entailment_checker_atom lambdas \<phi> \<psi>)"
  shows "\<phi> implies Conjunction [\<psi>]"
  unfolding entailment_def
proof (intro allI impI)
  obtain atoms where \<phi>: "\<phi> = Conjunction atoms" by (cases \<phi>, auto)
  def exps \<equiv> "map lin_exp atoms" 
  fix \<alpha>
  assume "eval_lin_formula \<alpha> \<phi>"
  hence exps: "\<And> e. e \<in> set exps \<Longrightarrow> eval_lin_exp \<alpha> e \<le> 0" unfolding exps_def \<phi> by auto
  let ?exp = "add_lin_exps (map (\<lambda> (c,e). scalar_mult (int c) e) (zip lambdas exps))"
  from ok[unfolded entailment_checker_atom_def Let_def]
  have ok: "isOK(trivial_entailment_checker ?exp (lin_exp \<psi>))" by (simp add: exps_def \<phi>)
  have "eval_lin_exp \<alpha> ?exp \<le> 0" unfolding add_lin_exps
  proof (rule listsum_le_0, auto)
    fix a b
    assume "(a, b) \<in> set (zip lambdas exps)"
    hence "b \<in> set exps" by (rule set_zip_rightD)
    from exps[OF this]
    show "int a * eval_lin_exp \<alpha> b \<le> 0" by (simp add: mult_nonneg_nonpos)
  qed
  with impliesD[OF trivial_entailment_checker[OF ok], of \<alpha>]
  show "eval_lin_formula \<alpha> (Conjunction [\<psi>])" unfolding exps_def \<phi> by auto
qed

datatype entailment_info = Entailment_Hint "nat list list"

fun entailment_checker :: "entailment_info \<Rightarrow> 'v formula \<Rightarrow> 'v formula \<Rightarrow> shows check" where
  "entailment_checker (Entailment_Hint hints) \<phi> (Conjunction psis) = do {
    check (length hints = length psis) (shows ''number of hints does not correspond to number of conjuncts in conclusion\<newline>'');
    check_allm (\<lambda> (h,\<psi>). entailment_checker_atom h \<phi> \<psi>) (zip hints psis)
  } <+? (\<lambda> s. shows ''could not ensure entailment of conjunction\<newline>  '' o 
      shows_formula \<phi> o shows '' ==> '' o shows_formula (Conjunction psis) o shows_nl o s
      )"

lemma entailment_conjunctionsI: "(\<And> \<psi>. \<psi> \<in> set psis \<Longrightarrow> \<phi> implies Conjunction [\<psi>]) \<Longrightarrow> \<phi> implies Conjunction psis"
  unfolding entailment_def by auto

lemma entailment_checker: assumes ok: "isOK(entailment_checker info \<phi> \<psi>)"
  shows "\<phi> implies \<psi>"
proof -
  obtain hints where info: "info = Entailment_Hint hints" by (cases info)
  obtain psis where \<psi>: "\<psi> = Conjunction psis" by (cases \<psi>, auto)
  note check = assms[unfolded \<psi> info]
  from check have len: "length hints = length psis" by auto
  show "\<phi> implies \<psi>" unfolding \<psi>
  proof (rule entailment_conjunctionsI)
    fix \<psi>
    assume "\<psi> \<in> set psis"
    with len obtain h where mem: "(h,\<psi>) \<in> set (zip hints psis)" 
      unfolding set_zip unfolding set_conv_nth by auto
    with check have "isOK(entailment_checker_atom h \<phi> \<psi>)" by auto
    from entailment_checker_atom[OF this]
    show "\<phi> implies Conjunction [\<psi>]" .
  qed
qed

end

end
