theory ITS
imports 
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
  "$AFP/Show/Show_Instances"
  "$AFP/Derive/Comparator_Generator/Compare_Generator"
begin

datatype 'v lin_exp = Lin_Exp int "('v \<times> int)list"

datatype 'v lin_atom = Exp_Le_0 (lin_exp: "'v lin_exp")

datatype 'v formula = Conjunction (atoms: "'v lin_atom list")

datatype 'v trans_var = Pre 'v | Post 'v | Intermediate 'v

derive compare_order trans_var

context
  assumes "SORT_CONSTRAINT('v :: show)"
begin
fun shows_lin_exp :: "'v lin_exp \<Rightarrow> shows" where
  "shows_lin_exp (Lin_Exp a cxs) = (shows a o foldr (\<lambda> (x,c) a. let ac = abs c;
    ope = (if c \<ge> 0 then '' + '' else '' - '');
    cc = (if ac = 1 then id else shows ac) in a o shows ope 
    o cc o shows x) cxs id)"

fun shows_lin_atom :: "'v lin_atom \<Rightarrow> shows" where
  "shows_lin_atom (Exp_Le_0 e) = (shows_lin_exp e o shows '' <= 0'' )"

fun shows_formula :: "'v formula \<Rightarrow> shows" where 
  "shows_formula (Conjunction phis) = shows_list_gen shows_lin_atom ''True'' [] '' && '' [] phis"
end

instantiation trans_var :: ("show") "show"
begin

fun shows_prec_trans_var :: "nat \<Rightarrow> 'a trans_var \<Rightarrow> shows" where
  "shows_prec_trans_var p (Pre x) = shows x"
| "shows_prec_trans_var p (Post x) = shows x o shows (Char Nibble2 Nibble7)"
| "shows_prec_trans_var p (Intermediate x) = shows x o  shows (Char Nibble2 Nibble7) o shows (Char Nibble2 Nibble7)"

lemma shows_prec_trans_var_append [show_law_simps]:
  "shows_prec p (v ::'a trans_var) (r @ s) = shows_prec p v r @ s"
  by (cases v, auto simp: show_law_simps)

definition "shows_list (fs :: 'a trans_var list) = showsp_list shows_prec 0 fs"

instance by standard (simp_all add: shows_list_trans_var_def show_law_simps)
end



type_synonym 'v transition_formula = "'v trans_var formula"

datatype ('v,'l)transition = Transition (trans_source: 'l) (trans_formula: "'v transition_formula") (trans_target: 'l)

type_synonym 'v valuation = "'v \<Rightarrow> int"

datatype ('v,'l)state = State "'v valuation" 'l

type_synonym ('v,'l)transitions = "('v,'l)transition set"

datatype ('v,'l,'t)lin_prog = Lin_Prog (initial: 'l) (transitions: "'t set") (transition_of: "'t \<Rightarrow> ('v,'l)transition")

datatype ('v,'l,'t)safety_prog = Safety_Prog 'l "('v,'l,'t)lin_prog"

context
fixes \<alpha> :: "'v valuation"
begin

fun eval_lin_exp :: "'v lin_exp \<Rightarrow> int" where
  "eval_lin_exp (Lin_Exp a cx) = a + (\<Sum>(x,c)\<leftarrow>cx. c * \<alpha> x)"

fun eval_lin_atom :: "'v lin_atom \<Rightarrow> bool" where
  "eval_lin_atom (Exp_Le_0 e) = (eval_lin_exp e \<le> 0)"

fun eval_lin_formula :: "'v formula \<Rightarrow> bool" where
  "eval_lin_formula (Conjunction phis) = (Ball (set phis) eval_lin_atom)"
end

fun pre_post_inter_valuation :: 
  "'v valuation \<Rightarrow> 'v valuation \<Rightarrow> 'v valuation \<Rightarrow> 'v trans_var valuation" where
  "pre_post_inter_valuation pre post inte (Pre x) = pre x"
| "pre_post_inter_valuation pre post inte (Post x) = post x"
| "pre_post_inter_valuation pre post inte (Intermediate x) = inte x"


context
fixes P :: "('v,'l,'t)lin_prog"
begin
definition transition :: "('v,'l)state rel" where
  "transition \<equiv> { (State \<alpha>1 l1, State \<alpha>2 l2) | l1 l2 \<alpha>1 \<alpha>2 \<phi> \<beta> \<tau>. 
    \<tau> \<in> transitions P \<and> transition_of P \<tau> = Transition l1 \<phi> l2 
    \<and> eval_lin_formula (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>) \<phi>}"

definition initial_states :: "('v,'l)state set" where
  "initial_states = { State \<alpha> l | \<alpha> l. l = initial P}"

definition "lin_prog_termination" :: bool where
  "lin_prog_termination \<equiv> SN_on transition initial_states"

definition reachable_states :: "('v,'l)state set" where
  "reachable_states = transition^* `` initial_states"
end

datatype 'l copy = Copy 'l | Cutpoint_Aux 'l | Error_Loc

datatype ('v,'l,'t)cooperation_prog = Cooperation_Prog 
  (orig_prog: "('v,'l,'t)lin_prog") 
  (cutpoints: "'l set")
  (deleted_transitions: "('v,'l copy)transitions")

fun copy_prog :: "('v,'l,'t)cooperation_prog \<Rightarrow> ('v,'l copy,'t)lin_prog" where
  "copy_prog P = undefined"

fun complete_copy_prog :: "('v,'l,'t)cooperation_prog \<Rightarrow> ('v,'l + 'l copy,'t)lin_prog" where
  "complete_copy_prog P = undefined"

context 
  fixes P :: "('v,'l,'t)cooperation_prog"
begin
definition cooperation_prog_termination :: bool where
  "cooperation_prog_termination \<equiv> SN_on (transition (copy_prog P)) 
    (map_state Copy ` reachable_states (orig_prog P))"
end

definition safety :: "('v,'l,'t)safety_prog \<Rightarrow> bool" where
  "safety SP \<equiv> case SP of Safety_Prog err P \<Rightarrow> (\<forall> \<alpha>. (State \<alpha> err) \<notin> reachable_states P)"

definition step1 :: "('v,'l,'t)lin_prog \<Rightarrow> ('v,'l,'t)cooperation_prog option" where
  "step1 = undefined"

lemma step1: "step1 P = Some Q \<Longrightarrow> cooperation_prog_termination Q \<Longrightarrow> lin_prog_termination P"
  oops

definition step2 :: "('v,'l,'t)cooperation_prog \<Rightarrow> ('v,'l,'t)cooperation_prog option" where
  "step2 = undefined"

lemma step2: "step2 P = Some Q \<Longrightarrow> cooperation_prog_termination Q \<Longrightarrow> cooperation_prog_termination P"
  oops

type_synonym rank_function_info = string

definition step3 :: "rank_function_info \<Rightarrow> ('v,'l,'t)cooperation_prog \<Rightarrow> ('w,'ll,'t)safety_prog option" where
  "step3 = undefined"

lemma step3: "step3 rfi P = Some SP \<Longrightarrow> safety SP \<Longrightarrow> cooperation_prog_termination P"
  oops
  
definition form_False :: "'v formula" where
  "form_False = Conjunction [Exp_Le_0 (Lin_Exp 1 [])]"

definition form_True :: "'v formula" where
  "form_True = Conjunction []"

lemma form_True[simp]: "eval_lin_formula \<alpha> form_True"
  by (simp add: form_True_def)

fun form_and :: "'v formula \<Rightarrow> 'v formula \<Rightarrow> 'v formula" (infix "and" 80) where
  "Conjunction \<phi> and Conjunction \<psi> = Conjunction (\<phi> @ \<psi>)"

lemma form_andI: "eval_lin_formula \<alpha> \<phi> \<Longrightarrow> eval_lin_formula \<alpha> \<psi> \<Longrightarrow> eval_lin_formula \<alpha> (\<phi> and \<psi>)"
  by (cases \<phi>; cases \<psi>; auto)

definition entailment :: "'v formula \<Rightarrow> 'v formula \<Rightarrow> bool" (infix "implies" 60) where
  "\<phi> implies \<psi> \<equiv> \<forall> \<alpha>. eval_lin_formula \<alpha> \<phi> \<longrightarrow> eval_lin_formula \<alpha> \<psi>"

lemma impliesD: "\<phi> implies \<psi> \<Longrightarrow> eval_lin_formula \<alpha> \<phi> \<Longrightarrow> eval_lin_formula \<alpha> \<psi>"
  unfolding entailment_def by auto

                                 
end
